#include "Panda.hpp"
#include "filesystem.h"
#include "mm_types.h"
#include "nds/arm9/background.h"
#include "nds/arm9/cache.h"
#include "nds/arm9/console.h"
#include "nds/arm9/input.h"
#include "nds/arm9/math.h"
#include "nds/arm9/sassert.h"
#include "nds/arm9/video.h"
#include "nds/cothread.h"
#include "nds/decompress.h"
#include "nds/dma.h"
#include "nds/interrupts.h"
#include "nds/timers.h"
#include "tonccpy.h"
#include <fat.h>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <nds.h>
#include <maxmod9.h>
#include <sys/_types.h>

#define PALETTE_SIZE 256

#define MMSTREAM_BUF_SIZE 800
#define MUSIC_RBUF_SIZE 16000

void wait_forever(const char* msg);
bool LoadNextChunk(unsigned char *dest);
void TimerCallback();
void VBLCallback();
void FrameStep();
int ThreadMusic(void *arg);
int ThreadVideo(void *arg);
mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format );

volatile bool doDraw = true;
volatile bool isPause = false;
volatile int noframe = 0;
volatile bool time_to_exit = false;



struct RingBuffer {
	unsigned char *buffer = nullptr;
	int head;
	int tail;
	size_t buffer_size;
	size_t size_per_item;
	size_t size;
	size_t count;

	RingBuffer(size_t count_frame, unsigned char * buffer_ptr, size_t buffer_len)
	{
		buffer = buffer_ptr;
		buffer_size = buffer_len;
		size = count_frame;
		head = 0;
		tail = 0;
		count = 0;
		size_per_item = buffer_size/size;
	}

	unsigned char* Get()
	{
		if (count > 0)
		{
			count--;
			unsigned char* ptr = &buffer[head*size_per_item];
			head = (head+1) % size;
			return ptr;
		}

		return nullptr;
	}

	unsigned char* Write()
	{
		if (count < size)
		{
			count++;
			unsigned char* ptr = &buffer[tail*size_per_item];
			tail = (tail+1) % size;
			return ptr;
		}

		return nullptr;
	}

	unsigned char* Manual_Read()
	{
		unsigned char* ptr = &buffer[head*size_per_item];
		return ptr;
	}

	unsigned char* Manual_Write()
	{
		unsigned char* ptr = &buffer[tail*size_per_item];
		return ptr;
	}

	void Manual_Skip(int total)
	{
		count += total;
		tail = (tail+total) % size;
	}

	void Manual_Advance(int total)
	{
		count -= total;
		head = (head+total) % size;
	}

	size_t getFree()
	{
		return size-count;
	}
};

struct Rbuf_Music {
	unsigned char *buffer = nullptr;
	int head;
	int tail;
	int size;
	int count;

	Rbuf_Music(size_t buffer_size, unsigned char* buffer_ptr)
	{
		buffer = buffer_ptr;
		size = buffer_size;
		head = 0;
		tail = 0;
		count = 0;
	}

	void Skip(int total)
	{
		count -= total;
		head = (head+total) % size;
	}

	void Advance(int total)
	{
		count += total;
		tail = (tail+total) % size;
	}

	int LinearReadSize()
	{
		return size - head;
	}

	int LinearWriteSize()
	{
		return size - tail;
	}

	size_t GetFree()
	{
		return size-count;
	}

	unsigned char *GetHead()
	{
		return &buffer[head];
	}

	unsigned char *GetTail()
	{
		return &buffer[tail];
	}

	int read(char *dest, size_t len)
	{
		if (count >= len)
		{
			count -= len;
			int linear_len = size - head;

			if (len < linear_len)
				memcpy(dest, &buffer[head], len);
			else
			{
				memcpy(dest, &buffer[head], linear_len);
				memcpy(&dest[linear_len], buffer, len-linear_len);
			}

			head = (head+len) % size;

			return 1;
		}

		return 0;
	}

	int write(char *data, size_t len)
	{
		if (count+len <= size)
		{
			count += len;
			int linear_len = size - tail;

			if (len < linear_len)
				memcpy(&buffer[tail], data, len);
			else
			{
				memcpy(&buffer[tail], data, linear_len);
				memcpy(buffer, &data[linear_len], len-linear_len);
			}

			tail = (tail+len) % size;

			return 1;
		}

		return 0;
	}
};

int ptr_background;
int ptr_subbackground;

FILE *file_music = nullptr;
int file_music_len;

FILE *file_videomain = nullptr;
int file_videomain_len;

uint8_t metadata_fps;
uint8_t metadata_frame_per_chunk;
uint16_t metadata_width1;
uint16_t metadata_width2;
uint16_t metadata_height1;
uint16_t metadata_height2;
uint32_t metadata_total_frame1;
uint32_t metadata_total_frame2;

uint32_t size_video1_frame;
uint32_t size_video2_frame;
uint32_t size_buffer_frame;

uint32_t counter_frame;

RingBuffer *videomain_rb;
unsigned char *frame_preload_temp;
unsigned char *frame_read_temp;

Rbuf_Music *music_rb;
unsigned char *music_buffer_temp;

uint32_t preload_video_size = 0;
uint32_t next_video_size = 0;

volatile int timer_start = 0;
volatile int timer_avg = 0;

PrintConsole console_top;
PrintConsole console_bot;

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_5_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankC(VRAM_C_SUB_BG_0x06200000);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	// consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	bool nitrofs = nitroFSInit(NULL);
	sassert(nitrofs, "nitrofs init failed");

	bool fatfs = fatInitDefault();
	sassert(fatfs, "fatfs init failed");

	file_videomain = fopen("nitro:/video.sillyvideo", "rb");
	sassert(file_videomain != nullptr, "failed to load video.sillyvideo");
	fseek(file_videomain, 0, SEEK_END);
	file_videomain_len = ftell(file_videomain);
	fseek(file_videomain, 0, SEEK_SET);

	file_music = fopen("nitro:/music1.raw", "rb");
	sassert(file_music != nullptr, "failed to load music1.raw");
	fseek(file_music, 0, SEEK_END);
	file_music_len = ftell(file_music);
	fseek(file_music, 0, SEEK_SET);

	ptr_background = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 1, 0);
	ptr_subbackground = bgInitSub(3, BgType_Bmp8, BgSize_B8_256x256, 1, 0);

	consoleInit(&console_top, 1, BgType_Text4bpp, BgSize_T_256x256, 2, 0, true, true);
	consoleInit(&console_bot, 1, BgType_Text4bpp, BgSize_T_256x256, 2, 0, false, true);

	bgSetPriority(ptr_background, 1);
	bgSetPriority(ptr_subbackground, 1);

	// consoleSetColor(&console, CONSOLE_RED);

	Panda::printf_nocash("scroll %i", metadata_height1);

	{
		Panda::printf_nocash("read metadata");
		fread(&metadata_fps, sizeof(uint8_t), 1, file_videomain);
		fread(&metadata_frame_per_chunk, sizeof(uint8_t), 1, file_videomain);
		fread(&metadata_total_frame1, sizeof(uint32_t), 1, file_videomain);
		fread(&metadata_total_frame2, sizeof(uint32_t), 1, file_videomain);
		fread(&metadata_width1, sizeof(uint16_t), 1, file_videomain);
		fread(&metadata_height1, sizeof(uint16_t), 1, file_videomain);
		fread(&metadata_width2, sizeof(uint16_t), 1, file_videomain);
		fread(&metadata_height2, sizeof(uint16_t), 1, file_videomain);

		size_video1_frame = metadata_width1*metadata_height1;
		size_video2_frame = metadata_width2*metadata_height2;
		size_buffer_frame = size_video1_frame + size_video2_frame;

		Panda::printf_nocash("fps: %i", metadata_fps);
		Panda::printf_nocash("frame per chunk: %i", metadata_frame_per_chunk);
		Panda::printf_nocash("video1 total frame: %i", metadata_total_frame1);
		Panda::printf_nocash("video2 total frame: %i", metadata_total_frame2);
		Panda::printf_nocash("video1 dimension: %ix%i", metadata_width1, metadata_height1);
		Panda::printf_nocash("video2 dimension: %ix%i", metadata_width2, metadata_height2);
	}

	bgSetScale(ptr_background, 160, 160);
	bgSetScale(ptr_subbackground, 160, 160);
	bgSetCenter(ptr_background, 0, (SCREEN_HEIGHT-144)/2);
	bgSetCenter(ptr_subbackground, 0, (SCREEN_HEIGHT-144)/2);
	bgUpdate();

	{
		Panda::printf_nocash("read palette");
		unsigned char *buffer_palette = (unsigned char*)sassert_malloc(PALETTE_SIZE);

		fread(buffer_palette, sizeof(uint8_t), PALETTE_SIZE*sizeof(uint16_t), file_videomain);
		DC_FlushRange(buffer_palette, PALETTE_SIZE*sizeof(uint16_t));
		dmaCopy(buffer_palette, BG_PALETTE, PALETTE_SIZE*sizeof(uint16_t));

		fread(buffer_palette, sizeof(uint8_t), PALETTE_SIZE*sizeof(uint16_t), file_videomain);
		DC_FlushRange(buffer_palette, PALETTE_SIZE*sizeof(uint16_t));
		dmaCopy(buffer_palette, BG_PALETTE_SUB, PALETTE_SIZE*sizeof(uint16_t));

		free(buffer_palette);
	}

	{
		Panda::printf_nocash("allocate buffer");

		frame_preload_temp = (unsigned char*)sassert_malloc(size_buffer_frame*metadata_frame_per_chunk);
		music_buffer_temp = (unsigned char*)sassert_malloc(MUSIC_RBUF_SIZE);

		int rbuf_count = 3145728 / (size_buffer_frame*metadata_frame_per_chunk);
		Panda::printf_nocash("rbuf %i", rbuf_count);

		size_t size_rbuf_count = rbuf_count*metadata_frame_per_chunk;
		size_t size_rbuf_buffer = size_rbuf_count*size_buffer_frame;
		
		unsigned char* buffer_video = (unsigned char*)sassert_malloc(size_rbuf_buffer);
		unsigned char* buffer_music = (unsigned char*)sassert_malloc(MUSIC_RBUF_SIZE);

		videomain_rb = new RingBuffer(rbuf_count*metadata_frame_per_chunk, buffer_video, size_rbuf_buffer);
		music_rb = new Rbuf_Music(MUSIC_RBUF_SIZE, buffer_music);

		memset(music_rb->buffer, 0, MUSIC_RBUF_SIZE);
	}

	// {
	// 	Raylib::nocashMessageFormat("calculate decompress time");

	// 	unsigned char *ptr = videomain_rb->Manual_Write();
	// 	FILE *data = fopen("fat:/numbers.txt", "w");

	// 	if (data == NULL) {
	// 		perror("Failed to open file");
  //   }

	// 	cpuStartTiming(1);

	// 	while (ftell(file_videomain) < file_videomain_len)
	// 	{
	// 		timer_start = cpuGetTiming();
	// 		fread(&preload_video_size, sizeof(uint32_t), 1, file_videomain);
	// 		fread(frame_preload_temp, sizeof(uint8_t), preload_video_size, file_videomain);
	// 		decompress(frame_preload_temp, ptr, DecompressType::LZ77);
	// 		timer_avg = timerTicks2msec(cpuGetTiming() - timer_start);
	// 		// time[counter] = timer_avg;

	// 		fprintf(stderr, "%i,\n", timer_avg);
	// 		fprintf(data, "%i, ", timer_avg);
	// 	}

	// 	fclose(data);
	// 	cpuEndTiming();
	// 	Raylib::nocashMessageFormat("done calculate");
	// 	fseek(file_videomain, 0, SEEK_SET);
	// }

	{
		Panda::printf_nocash("preload video main");
		do {
			unsigned char *ptr = videomain_rb->Manual_Write();
			LoadNextChunk(ptr);
			videomain_rb->Manual_Skip(metadata_frame_per_chunk);
		} while(videomain_rb->getFree() >= metadata_frame_per_chunk);

		Panda::printf_nocash("preload done");
	}

	{
		Panda::printf_nocash("preload audio");
		int len = music_rb->GetFree();
		int adv = fread(music_rb->GetTail(), sizeof(char), len, file_music);
		music_rb->Advance(adv);
		Panda::printf_nocash("preload done");
	}

	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			  = 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit(&sys);

	mm_stream mystream;
	mystream.sampling_rate	= 16000;
	mystream.buffer_length	= MMSTREAM_BUF_SIZE;
	mystream.callback		= on_stream_request;
	mystream.format			= MM_STREAM_8BIT_MONO;
	mystream.timer			= MM_TIMER1;
	mystream.manual			= false;

	timerStart(0, ClockDivider_1024, TIMER_FREQ_1024(metadata_fps), TimerCallback);
	irqSet(IRQ_VBLANK, VBLCallback);

	// cothread_create(ThreadMusic, NULL, 0, COTHREAD_DETACHED);
	// cothread_create(ThreadVideo, NULL, 0, COTHREAD_DETACHED);

	// cpuStartTiming(1);

	mmStreamOpen( &mystream );

	while(!time_to_exit) {
		// swiWaitForVBlank();
		// cothread_yield_irq(IRQ_VBLANK);
		// scanKeys();


		// if (keysUp() & KEY_A) FrameStep();

		if (!isPause)
		{
			if (file_videomain)
			{
				unsigned char *ptr = videomain_rb->Manual_Write();
				if (videomain_rb->getFree() >= metadata_frame_per_chunk)
				{
					// timer_start = cpuGetTiming();
					LoadNextChunk(ptr);
					videomain_rb->Manual_Skip(metadata_frame_per_chunk);
					// timer_avg = cpuGetTiming() - timer_start;
				}
			}

			if (ftell(file_videomain) == file_videomain_len)
			{
				fclose(file_videomain);
				file_videomain = nullptr;
				// return 0;
			}

			if (file_music != nullptr && music_rb->count < music_rb->size*0.7)
			{
				int free = music_rb->GetFree();
				int linear = music_rb->LinearWriteSize();
				int adv = 0;

				// memset(music_buffer_temp, 0, MUSIC_RBUF_SIZE);
				// fread(music_buffer_temp, free, 1, file_music);


				if (free < linear)
				{
					// memcpy(music_rb->GetTail(), music_buffer_temp, free);
					adv += fread(music_rb->GetTail(), 1, free, file_music);
				}
				else
				{
					// memcpy(music_rb->GetTail(), music_buffer_temp, linear);
					// memcpy(music_rb->buffer, music_buffer_temp, free-linear);
					adv += fread(music_rb->GetTail(), 1, linear, file_music);
					adv += fread(music_rb->buffer, 1, free-linear, file_music);
				}

				music_rb->Advance(adv);

			}

			if (ftell(file_music) >= file_music_len)
			{
				fclose(file_music);
				file_music = nullptr;
				Panda::printf_nocash("music close!");
			}

			static bool closeOnce = false;

			if (!closeOnce && file_music == nullptr && music_rb->count <= 0)
			{
				mmStreamClose();
				Panda::printf_nocash("stream close!");
				closeOnce = true;
				time_to_exit = true;
			}
		}

	}

	return 0;
}

void VBLCallback()
{
	static int counter = 0;
	counter++;

	if (counter % 8 == 0) {

		consoleSelect(&console_top);
		// float progress = counter_frame/metadata_total_frame1;
		Panda::f32 f32_counter(inttof32(counter_frame));
		Panda::f32 f32_total(inttof32(metadata_total_frame1));
		Panda::f32 f32_progress((f32_counter/f32_total) * Panda::f32(inttof32(100)));
		Panda::f32 f32_bar((f32_counter/f32_total) * Panda::f32(inttof32(30)));

		printf("\033[22;1H%i%% (%i/%i)\n[", f32toint(f32_progress.value), counter_frame, metadata_total_frame1);
		for(int i=0; i<f32toint(f32_bar.value); i++) printf("-");
		printf("\033[23;32H]");

		consoleSelect(&console_bot);
		printf("\033[21;1H\033[Kvideo buffer: %i frames", videomain_rb->count);
		printf("\033[22;1H\033[Kmusic buffer: %i bytes", music_rb->count);
		if (noframe) printf("\033[23;1H\033[Kframe drop: %i", noframe);
		// else printf("\033[23;1H\033[Kframe drop: %i yay \\ >.< /\n", noframe);
	}
	
	// 	printf("\033[6;2H\033[Kavg decode time: %ims\n", timerTicks2msec(timer_avg/30));

	scanKeys();
	if (keysUp() & KEY_A) isPause = !isPause;
	if (keysUp() & KEY_B) FrameStep();
	if (keysHeld() & KEY_Y) FrameStep();
}

void FrameStep()
{
	if (videomain_rb->count == 0 && !file_videomain) return;
	
	if (videomain_rb->count)
	{
		unsigned char *dest = videomain_rb->Get();
		dmaCopyAsynch(dest, bgGetGfxPtr(ptr_background), size_video1_frame);
		dmaCopyAsynch(dest+size_video1_frame, bgGetGfxPtr(ptr_subbackground), size_video2_frame);
		counter_frame++;
	} else {
		// printf("no frame!!!!!\n");
		noframe++;
	}
	// }
}

void TimerCallback()
{
	FrameStep();
	doDraw = false;

	// printf("\033[4;2H\033[Kvideo buffer: %i", videomain_rb->count/(FRAME_SIZE));
	// printf("\033[5;2H\033[Kmusic buffer: %i", music_rb->count/(MMSTREAM_BUF_SIZE*4));
}

bool LoadNextChunk(unsigned char *dest)
{
	if (!dest) return false;

	fread(&preload_video_size, sizeof(uint32_t), 1, file_videomain);
	// Panda::printf_nocash("load next start, %i", preload_video_size);
	if (preload_video_size == size_buffer_frame*metadata_frame_per_chunk)
	{
		fread(dest, sizeof(uint8_t), preload_video_size, file_videomain);
	}
	else
	{
		fread(frame_preload_temp, sizeof(uint8_t), preload_video_size, file_videomain);
		decompress(frame_preload_temp, dest, DecompressType::LZ77);
	}
	// fastlz_decompress(frame_preload_temp, preload_video_size, dest, FRAME_SIZE*CHUNK_SIZE);

	return true;
}

mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format ) {
	size_t samplesize = 1;
	switch(format){
		case MM_STREAM_8BIT_MONO: samplesize = 1; break;
		case MM_STREAM_8BIT_STEREO: samplesize = 2; break;
		case MM_STREAM_16BIT_MONO: samplesize = 2; break;
		case MM_STREAM_16BIT_STEREO: samplesize = 4; break;
	}
	int len = length*samplesize;
	int linear = music_rb->LinearReadSize();

	// Raylib::nocashMessageFormat("%i", len);
	
	// if (ftell(file_music) >= file_music_len) mmStreamClose();
	// fread(music_buffer_temp, 1, len, file_music);
	// memcpy(dest, music_buffer_temp, len);
	// return length;

	// if (len < linear)
	// {
	// 	memcpy(dest, music_rb->GetHead(), len);
	// 	music_rb->Skip(len);
	// 	return length;
	// }
	// else
	// {
	// 	// Raylib::nocashMessageFormat("linear %i", linear);
	// 	memcpy(dest, music_rb->GetHead(), linear);
	// 	// memcpy((char*)dest+linear, music_rb->buffer, len-linear);
	// 	music_rb->Skip(linear);
	// 	return linear/samplesize;
	// }


	if (len < linear)
	{
		tonccpy(dest, music_rb->GetHead(), len);
	}
	else
	{
		// Raylib::nocashMessageFormat("linear %i", linear);
		tonccpy(dest, music_rb->GetHead(), linear);
		tonccpy((char*)dest+linear, music_rb->buffer, len-linear);
	}

	music_rb->Skip(len);
	return length;
}